from __future__ import annotations

from io import StringIO, BytesIO
from pathlib import Path

from ._types import FileData, FileType


def write_file(data: FileData, file: FileType, mode: str, fake=False):
    if isinstance(file, str):
        file = Path(file)

    if fake:
        if "w" in mode:
            print(f"fake ❯ write in '{file}'")
        else:
            print(f"fake ❯ append to '{file}'")
    elif isinstance(file, Path):
        if not file.parent.exists():
            raise ValueError(f"Can't find the directory '{file.parent}'")

        mode = f"{mode}b" if isinstance(data, bytes) else mode
        with open(file, mode) as f:
            f.write(data)
    elif isinstance(file, StringIO):
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        file.write(data)
    elif isinstance(file, BytesIO):
        if isinstance(data, str):
            data = data.encode("utf-8")
        file.write(data)
    else:
        raise ValueError(f"Wrong file type '{file}'")



from _typeshed import Incomplete
from collections import UserDict

ALIASES: Incomplete

class Aliases(UserDict):
    @classmethod
    def default(cls): ...
    @staticmethod
    def compile(alias): ...
    def process_aliases(self, cmd: str): ...
    regexes: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...
    def __setitem__(self, alias, replacement) -> None: ...
    def __delitem__(self, alias) -> None: ...



from ._types import NotSet as NotSet, RunPipeCmd as RunPipeCmd, not_set as not_set
from .shell import Shell as Shell
from typing import Any, Iterable

class XArgsMeta(type):
    def __sub__(self, cmd: RunPipeCmd) -> XArgs: ...
    def __call__(self, *args, **kwargs) -> XArgs: ...

class XArgs(metaclass=XArgsMeta):
    stdin: Any
    execute: Union[RunPipeCmd, NotSet]
    def __call__(self, stdin) -> XArgs: ...
    def __sub__(self, cmd: RunPipeCmd) -> XArgs: ...
    def iter_stdin(self) -> Iterable: ...
    def run(self, cmd: RunPipeCmd, shell: Shell) -> list: ...
    def __init__(self, stdin, execute) -> None: ...
xargs = XArgs



from io import StringIO, BytesIO
from pathlib import Path
from unittest.mock import Mock

import pytest

from bulot.file_operations import write_file
from .constants import ECHO_STDOUT, FILE_NAME, PATH_TESTS


class TestFileOperations:
    def test_wrong_directory(self, file_mocker: Mock):
        with pytest.raises(ValueError):
            write_file(ECHO_STDOUT, PATH_TESTS / "plop" / FILE_NAME, "w")

    @pytest.mark.usefixtures("file_mocker")
    def test_write(self, file_mocker: Mock):
        write_file(ECHO_STDOUT, Path(FILE_NAME), "w")
        file_mocker.assert_called_once_with(Path(FILE_NAME), "w")
        file_mocker().write.assert_called_once_with(ECHO_STDOUT)

    @pytest.mark.usefixtures("file_mocker")
    def test_append(self, file_mocker: Mock):
        write_file(ECHO_STDOUT, FILE_NAME, "a+")
        file_mocker.assert_called_once_with(Path(FILE_NAME), "a+")
        file_mocker().write.assert_called_once_with(ECHO_STDOUT)

    @pytest.mark.usefixtures("file_mocker")
    def test_write_stringio(self, file_mocker: Mock):
        file = StringIO()
        write_file(ECHO_STDOUT, file, "w")
        assert file.getvalue() == ECHO_STDOUT

    @pytest.mark.usefixtures("file_mocker")
    def test_write_stringio_str(self, file_mocker: Mock):
        file = StringIO()
        data = ECHO_STDOUT.encode("utf-8")
        write_file(data, file, "w")
        assert file.getvalue() == ECHO_STDOUT

    @pytest.mark.usefixtures("file_mocker")
    def test_write_bytesio(self, file_mocker: Mock):
        file = BytesIO()
        data = ECHO_STDOUT.encode("utf-8")
        write_file(data, file, "w")
        assert file.getvalue() == data

    @pytest.mark.usefixtures("file_mocker")
    def test_write_bytesio_str(self, file_mocker: Mock):
        file = BytesIO()
        data = ECHO_STDOUT.encode("utf-8")
        write_file(ECHO_STDOUT, file, "w")
        assert file.getvalue() == data



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .pipe import Pipe


class PipeExtractor:
    attr: str

    def __call__(self, pipe: "Pipe") -> Any:
        return getattr(pipe, self.attr)


class Stdout(PipeExtractor):
    attr = "stdout"


class Stderr(PipeExtractor):
    attr = "stderr"


class Stdin(PipeExtractor):
    attr = "stdin"


stdout = Stdout()
stderr = Stderr()
stdin = Stdin()



from __future__ import annotations

import subprocess as sp
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from .file_operations import write_file
from ._types import FileType


@dataclass
class Command:
    value: str
    cwd: Path
    start: float
    end: float
    return_code: int
    result: None | sp.CalledProcessError | sp.CompletedProcess
    fake: bool = False

    @property
    def stdout(self) -> str:
        if self.result is None:
            return ""
        else:
            stdout = self.result.stdout.decode("utf-8")
            return Command._remove_trailling_new_line(stdout)

    @property
    def stderr(self) -> str:
        if self.result is None:
            return ""
        else:
            stderr = self.result.stderr.decode("utf-8")
            return Command._remove_trailling_new_line(stderr)

    @staticmethod
    def _remove_trailling_new_line(value: str) -> str:
        if value.endswith("\n"):
            return value[:-1]
        else:
            return value

    def __gt__(self, file: FileType) -> Any:
        write_file(self.stdout, file, "w", self.fake)
        return self.stdout

    def __rshift__(self, file: FileType) -> Any:
        write_file(self.stdout, file, "a+", self.fake)
        return self.stdout



from pathlib import Path
from unittest.mock import Mock

import pytest

from bulot import Pipe, Shell, xargs, XArgs
from .constants import FILE_NAME, ECHO_CMD, ECHO_STDOUT, PIPE_STDOUT, REPLACE_STDOUT


class TestPipe:
    pipe = Pipe(Shell(), PIPE_STDOUT)

    def test__repr__(self):
        assert repr(self.pipe) == f"Pipe(stdout={repr(PIPE_STDOUT)})"

    def test__or__str(self):
        pipe = self.pipe | f"sed 's/{PIPE_STDOUT}/{REPLACE_STDOUT}/'"
        assert pipe.stdout == REPLACE_STDOUT

    def test__or__callable(self):
        def replace(old: str) -> str:
            return old.replace(PIPE_STDOUT, REPLACE_STDOUT)

        pipe = self.pipe | replace
        assert pipe.stdout == REPLACE_STDOUT

    def test__or__xargs(self):
        pipe = self.pipe | xargs
        x = pipe.stdout
        assert isinstance(x, XArgs)
        assert x.stdin == PIPE_STDOUT

    @pytest.mark.usefixtures("file_mocker")
    def test__gt__(self, file_mocker: Mock):
        self.pipe.sh | ECHO_CMD > FILE_NAME
        file_mocker.assert_called_once_with(Path(FILE_NAME), "w")
        file_mocker().write.assert_called_once_with(ECHO_STDOUT)

    @pytest.mark.usefixtures("file_mocker")
    def test__rshift__(self, file_mocker: Mock):
        (self.pipe.sh | ECHO_CMD) >> FILE_NAME
        file_mocker.assert_called_once_with(Path(FILE_NAME), "a+")
        file_mocker().write.assert_called_once_with(ECHO_STDOUT)



from ._py_version import PY as PY
from ._types import RunPipeCmd as RunPipeCmd
from .aliases import Aliases as Aliases
from .command import Command as Command
from .logger import log as log
from .pipe import Pipe as Pipe
from _typeshed import Incomplete
from pathlib import Path
from typing import Any, Callable

class RunConfig:
    input: Union[str, bytes, None]
    capture_output: bool
    timeout: Union[int, None]
    check: bool
    bufsize: int
    executable: Union[str, None]
    stdin: Union[int, None]
    stdout: Union[int, None]
    stderr: Union[int, None]
    preexec_fn: Union[Callable, None]
    close_fds: bool
    shell: bool
    cwd: Union[str, None]
    env: Union[dict, None]
    universal_newlines: None
    startupinfo: None
    creationflags: int
    restore_signals: bool
    start_new_session: bool
    pass_fds: tuple
    encoding: None
    errors: None
    text: None
    @classmethod
    def default(cls) -> RunConfig: ...
    def __init__(self, input, capture_output, timeout, check, bufsize, executable, stdin, stdout, stderr, preexec_fn, close_fds, shell, cwd, env, universal_newlines, startupinfo, creationflags, restore_signals, start_new_session, pass_fds, encoding, errors, text) -> None: ...

def do_nothing(*args, **kwargs) -> None: ...

class Shell:
    fake: bool
    config: RunConfig
    historic: list[Command]
    silent_piping: bool
    aliases: Incomplete
    def __post_init__(self, *args, **kwargs) -> None: ...
    def cd(self, path: Union[Path, str]): ...
    def cwd(self) -> Path: ...
    def run(self, cmd: str, *args, fake: bool = ..., callback=..., silent: bool = ..., **kwargs) -> Command: ...
    def __call__(self, *args, **kwargs): ...
    def __or__(self, cmd: RunPipeCmd) -> Pipe: ...
    def __lshift__(self, stdin: Any) -> Pipe: ...
    def run_pipe(self, cmd: RunPipeCmd, **kwargs) -> Union[Any, Command]: ...
    def __init__(self, _cwd, fake, config, historic, silent_piping) -> None: ...



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import Any, Iterable, TYPE_CHECKING


from ._types import not_set, NotSet, RunPipeCmd

if TYPE_CHECKING:
    from .shell import Shell


class XArgsMeta(type):
    def __sub__(self, cmd: RunPipeCmd) -> "XArgs":
        return self(execute=cmd)

    def __call__(self, *args, **kwargs) -> "XArgs":
        return super().__call__(*args, **kwargs)


@dataclass
class XArgs(metaclass=XArgsMeta):
    stdin: Any = not_set
    execute: RunPipeCmd | NotSet = not_set

    def __str__(self) -> str:
        return str(self.stdin)

    def __call__(self, stdin) -> XArgs:
        self.stdin = stdin
        return self

    def __sub__(self, cmd: RunPipeCmd) -> XArgs:
        ex = self.execute
        if ex is not_set:
            ex = cmd
        else:

            def chain(cmd, ex, *args, **kwargs) -> Any:
                return cmd(ex(*args, **kwargs))

            ex = partial(chain, cmd, ex)

        self.execute = ex
        return self

    def iter_stdin(self) -> Iterable:
        stdin = self.stdin
        if isinstance(stdin, str):
            return stdin.split()
        elif isinstance(stdin, list):
            return stdin
        else:
            raise NotImplementedError(f"XArgs is not implemented for {type(stdin)}")

    def run(self, cmd: RunPipeCmd, shell: Shell) -> list:
        it = self.iter_stdin()

        execute = self.execute
        if isinstance(execute, NotSet):
            stdout = [shell.run_pipe(cmd, input=stdin).stdout for stdin in it]
        else:
            if isinstance(execute, str):
                execute = execute.format
                stdout = [shell.run_pipe(execute(stdin)).stdout for stdin in it]
            else:
                stdout = [execute(stdin) for stdin in it]

        return stdout


xargs = XArgs



import re

from bulot.aliases import Aliases, ALIASES


# To provide correct tests we must be sure that ALIASES still caontains enough entries
assert len(ALIASES) > 1


class TestAliases:
    aliases = Aliases(ALIASES)

    def test_regex(self):
        for alias, (regex, replacement) in self.aliases.regexes.items():
            assert isinstance(alias, str)
            assert isinstance(regex, re.Pattern)
            assert isinstance(replacement, str)

    def test_process(self):
        cmd = " ".join(ALIASES.keys())
        cmd = self.aliases.process_aliases(cmd)
        assert cmd == " ".join(ALIASES.values())



from .xargs import XArgs as XArgs, XArgsMeta as XArgsMeta
from _typeshed import Incomplete
from io import BytesIO, StringIO
from pathlib import Path
from typing import Callable, Union

class NotSet: ...
RunPipeCmd = Union[str, Callable]
PipeCmd: Incomplete
not_set: Incomplete
FileData = Union[str, bytes]
FileType = Union[str, Path, StringIO, BytesIO]



from ._types import FileType as FileType, PipeCmd as PipeCmd
from .command import Command as Command
from .file_operations import write_file as write_file
from .shell import Shell as Shell
from .utils import PipeExtractor as PipeExtractor
from .xargs import XArgs as XArgs, XArgsMeta as XArgsMeta
from typing import Any

class Pipe:
    sh: Shell
    stdout: Any
    stderr: Union[Any, None]
    return_code: int
    def __or__(self, cmd: Union[PipeCmd, PipeExtractor]) -> Union[Pipe, Any]: ...
    def __gt__(self, file: FileType) -> Any: ...
    def __rshift__(self, file: FileType) -> Any: ...
    def __init__(self, sh, stdout, stderr, return_code) -> None: ...
